part of tags;

/// A function to build children of a [Tag].
///
/// Each tag function that is called from within a build function will append
/// its element as a child of the current [buildHostNode].  The order of such 
/// calls is equal to the order in which the children are appended.
typedef void BuildChildren();

/// A function to create an instance of a node with the given [args].
///
/// All tag functions called from within the scope of this function are inert
/// and will _not_ append any nodes.  The [buildHostNode] is always _null_ 
/// within the scope of this function.  This allows the side-effect free usage
/// of tags in class initializers.
typedef CreateNode(args);

/// A function to build an html element.
///
/// An optional [build] function may be provided to build child elements.
/// 
/// An optional [args] may be provided and will be applied to the [CreateNode] 
/// function call for this tag.
typedef Tag([BuildChildren build, args]);

// Stack of nodes used to build trees.  
// The top of the stack is always the host for the current build scope.
final _scopes = new List();

// Whether or not the builder is currently in a [Build] function scope.
bool _isBuilding = false;

/// The node currently being built; will be `null` if none or within the scope
/// of a [CreateNode] function.
get buildHostNode {
  if (!_isBuilding || _scopes.isEmpty) {
    return null;
  } else {
    return _scopes.last;
  }
}

/// Append the given [node] to the current [buildHostNode].
///
/// An optional [build] function may be provided to build child elements.  
/// 
/// This function provides a way to append existing nodes within the scope of a 
/// [BuildChildren] function.
append(node, [BuildChildren build]) => _buildNode((_) => node, build);

/// Append the given [text] as a text node to the current [buildHostNode].
///
/// This function may only be called from inside a [BuildChildren] function and 
/// the [buildHostNode] must be an element.
void appendText(String text) {
  final _host = buildHostNode;
  assert(_host != null);
  _host.appendText(text);
}

/// Create a node and build its children.
///
/// The node constructed by [createNode] will be the [buildHostNode] when the 
/// [buildChildren] function is called.  All [Tag] functions that are called 
/// within the scope of [buildChildren] will append their element as a child of 
/// the [buildHostNode].  [Tag] functions called within the scope of 
/// [createNode] are side-effect free.
_buildNode(CreateNode createNode, [BuildChildren buildChildren, args]) {
  final bool outerIsBuilding = _isBuilding;
  var node;
  try {
    _isBuilding = false;
    node = createNode(args);
    assert(node != null);
  } finally {  
    _isBuilding = outerIsBuilding;  
  }
  if (outerIsBuilding && !_scopes.isEmpty) {
    _scopes.last.append(node);  
  }
  if (buildChildren == null) {
    return node;
  }
  _scopes.add(node);  
  try {
    _isBuilding = true;
    buildChildren();
  } finally {
    _isBuilding = outerIsBuilding;
    _scopes.removeLast();
  }
  return node;
}

/// Build the children of the given [parent] element.
children(parent, BuildChildren buildChildren) =>
    _buildNode((_) => parent, buildChildren);

/// Creates a new [Tag] with the given [createNode] function.
Tag createTag(CreateNode createNode) => 
    ([BuildChildren b, args]) => _buildNode(createNode, b, args);

/// Create a shadow root for the current [buildHostNode].
///
/// This function may only be called from inside a [BuildChildren] function and 
/// the host must be an element.
shadow([BuildChildren b]) {
  final _host = buildHostNode;
  assert(_host != null);
  return _buildNode((_) => _host.createShadowRoot(), b);
}
