part of tags.html;

class TreeViewResources extends Resources {
  String get style =>
"""
.tree-item {
  color: ${theme.colors.text};
  display: flex;
  align-items: center;
  font-family: ${theme.display.fontFamily};
  font-size: ${theme.display.fontSize};
  height: ${theme.display.boxSize};
  width: calc(100% - 5px);
  margin: 0px 1px 1px 0px;
  padding: 0px 2px;
}
.tree-item.light {
  background-color: ${theme.colors.backgroundLight};
}
.tree-item.active {
  background-color: ${theme.colors.selection};
}
.tree-item:hover {
  background-color: ${theme.colors.hover};
}
.tree-item:hover * {
  color: ${theme.colors.textHover} !important;
}
.tree-expander {
  background-color: transparent;
  border: none;
  color: ${theme.colors.text};
  margin: 0;
  padding: 0;
  height: 100%;
  width: calc(${theme.display.boxSize} - 2px);
}
.tree-expander:focus {
  outline: none;
}
.tree-cell {
  color: ${theme.colors.text};
  display: flex;
  align-items: center;
  overflow: hidden;  
  white-space: nowrap;
  text-overflow: ellipsis;
}
.tree-cellbox.expandable {
  width: calc(100% - ${theme.display.boxSize} + 2px);
}
"""
// TODO: Investigate alternatives for rendering child tree list.
"""
table.tree-child-list {
  border-collapse: collapse;
  border-left: 1px solid ${theme.colors.backgroundLight};
  margin: 0px;    
  padding: 0px;
  position: relative;
  left: ${theme.display.boxSize};
  table-layout: fixed;
  width: calc(100% - ${theme.display.boxSize});
}
td.tree-child {
  color: ${theme.colors.text};
  vertical-align: middle;
  margin: 0px;
  overflow: hidden;
  padding: 0px;
  text-overflow: ellipsis;
}
""";
  const TreeViewResources();
}

abstract class TreeView<T> extends DivElement with Themeable
    implements Tree<T> {  
  
  T _data;
  T get data => _data;
  
  Tree _logicalParent;
  Tree get logicalParent => _logicalParent;
  
  final List<Tree> logicalChildren = new List<Tree>();
    
  bool _isExpanded = false;
  bool get isExpanded => _isExpanded;
    set isExpanded(bool value) {
      if (value != _isExpanded) {
        _isExpanded = value;
        if (_expander != null) {
          _expander.isDown = _isExpanded;
        }
      }
    }
  
  bool _isSelected = false;
  
  final TreeViewResources resources = const TreeViewResources();
  StyleElement styleElement;
  
  ToggleButton _expander;  
  Element _item;
  Element _cell;
    
  StreamSubscription _selectionListener;
  StreamSubscription _themeListener;
  
  TreeView.created() : super.created();

  void initialize(T data, [Tree logicalParent]) {
    _data = data;
    _logicalParent = logicalParent;
    initialize_(data);
    
    children(this, () {
      
      // Inject style only if this element is a tree root.
      if (logicalParent == null) {
        styleElement = new StyleElement();
        append(styleElement)..text = resources.style;
      }
     
      append(new CommandBinding(routedCommands[#copy], copy));
      append(new CommandBinding(routedCommands[#cut], cut));
      append(new CommandBinding(routedCommands[#delete], delete));
      append(new CommandBinding(routedCommands[#paste], paste));
      append(new CommandBinding(routedCommands[#rename], rename));
            
      _item = span(() {
        _expander = toggleButton(() {
          span()
          ..innerHtml = '&#9655;'
          ..classes = ['up'];
          span()
          ..innerHtml = '&#9698;'
          ..classes = ['down'];
        })
        ..classes = ['tree-expander']
        ..onToggle.listen((bool isDown) {
          _isExpanded = isDown;
          if (isDown) {
            expand_();
          } else {
            collapse_();
          }
          renderChildren();
        })
        ..isDown = isExpanded;
        _cell = span()
        ..classes.add('tree-cell up down');
        render(_cell);
      })
      ..classes.add('tree-item')
      ..onClick.listen((MouseEvent ev) {        
        if (!_expander.contains(ev.target)) {
          singleClick_();
        }
        _select(fireEvent: true);
      })
      ..onDoubleClick.listen((MouseEvent ev) {
        if (!_expander.contains(ev.target)) {
          doubleClick_();
        }
      })
      ..onContextMenu.listen((_) => showContextMenu_())
      ..onMouseOut.listen((_) => hideContextMenu_());
      
      table(() => _childList = tbody())
      ..classes.add('tree-child-list');
    });
    
    _selectionListener = Selections.onSelectionChange.listen((Selections s) {
      if (s.objects.contains(this.data)) {
        _select();
      } else {
        _unselect();
      }
    });
    
    if (logicalParent == null) {
      _themeListener = onThemeChange.listen((_) => refreshStyle());  
    }
    
    _renderExpander();
  }
    
  void refresh() {
    render(_cell);
    renderChildren();
  }
  
  void removeChildren(bool predicate(Tree tree)) {
    for (Tree child in logicalChildren.where(predicate)) {
      child.shutdown();
    }
    logicalChildren.removeWhere(predicate);
  }
  
  void render(Element parent) {
    parent.text = data.toString();
  }
  
  void renderChildren() {
    if (isExpanded) {
      _renderChildTrees(logicalChildren);
    } else {
      _renderChildTrees([]);
    }
    _renderExpander();
  }
  
  void _renderExpander() {
    if (logicalChildren.isEmpty) {
      _expander.hidden = true;
      _cell.classes.remove('expandable');      
    } else {
      _expander.hidden = false;
      _cell.classes.add('expandable');
    }
  }
  
  void _select({bool fireEvent: false}) {
    if (!_isSelected) {
      _isSelected = true;      
      _item.classes.add('active');
      scrollIntoView();
      _item.focus();
      select_();
      if (data != null && fireEvent) {
        Selections.select(data); 
      }
    }
  }  
  
  void toggleExpanded() {
    isExpanded = !isExpanded;
  }
      
  void _unselect() {
    if (_isSelected) {
      _isSelected = false;
      _item.classes.remove('active');
      unselect_(); 
    }
  }
      
  void attached() {
    super.attached();    
    attached_();
  }
  
  void detached() {
    super.detached();
    detached_();
  }
  
  void shutdown() {
    logicalChildren.forEach((Tree child) => child.shutdown());
    shutdown_();
    _selectionListener.cancel();
    if (_themeListener != null) {
      _themeListener.cancel();
    }
    _childCache.clear();
  }
  
// -----------------------------------------------------------------------------
// Child tree rendering.
// -----------------------------------------------------------------------------
  
  final _childCache = new Map<Tree, TableRowElement>.identity();
  Element _childList;
  
  void _renderChildTrees(List<Tree> trees) {
    for (Tree t in _childCache.keys.where((t) => 
        !trees.any((e) => identical(e, t)))) {
      _childList.children.remove(_childCache[t]);
    }
    for (int i = 0; i < trees.length; i++) {
      final Tree t = trees[i];            
      if (_childCache.containsKey(t)) {
        final Element el = _childCache[t];
        _insertChildAt(el, i);
        final TreeView tv = el.firstChild.firstChild;
        tv.refresh();
      } else {
        final el = tr(() => 
            td(() => append(t as TreeView))..classes.add('tree-child'));
        _childCache[t] = el;
        _insertChildAt(el, i);
      }
    }
  }
  
  void _insertChildAt(Element element, int index) {
    if (index < _childList.children.length) {
      _childList.insertBefore(element, _childList.children[index]);
    } else {
      _childList.append(element);
    }
  }
  
// -----------------------------------------------------------------------------
// Tree stubs; concrete TreeViews should mixin an implementation.
// -----------------------------------------------------------------------------
  
  final Command copy = Command.noop;
  final Command cut = Command.noop;
  final Command delete = Command.noop;
  final Command paste = Command.noop;
  final Command rename = Command.noop;  
  
  void showContextMenu_() {}
  void hideContextMenu_() {}
  void singleClick_() {}
  void doubleClick_() {}
  void collapse_() {}
  void expand_() {}
  void attached_() {}
  void detached_() {}
  void initialize_(T data) {}
  void shutdown_() {}  
  void select_() {}
  void unselect_() {}
}
